#define STRICT      // Strict type checking
#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include "kxmenuid.h"	// Menu identifiers
#include "kxglobal.h"   // Global variables
#include "kxwproto.h"	// Function Prototypes


static int BeginRow, EndRow, BeginCol, EndCol, FontLead;
//#define _DETECT_BUGS_
#ifdef _DETECT_BUGS_
static int BUG = 0;
static LPSTR BugMsg= "Congradulations!\nYou have found a bug.\n\
The bug's ID is BUG%d.\n\Please report this to the author.\n\
\ne-mail: chenxj@cs.ucdavis.edu\n\nThank you for your help.";

int PASCAL CheckBug( TextWin *TW, int id, int Bug )
{
	if ( Bug ) {
		ValidateRect( TW->hWnd, NULL );
		if ( !BUG ) {
			BUG = 1;
			wsprintf( CommonBuffer, BugMsg, id );
			MessageBox( ThisApphWnd, CommonBuffer, "BUG!!!",
				MB_OK|MB_ICONHAND	 );
		}
		return 1;
	}
	return 0;
}

#endif



//
// Refresh the display by the encoding scheme
// 
void PASCAL RefreshScreen( TextWin *TW )
{
	if ( TW->ReloadBuffer ) LoadToScrBuffer( TW );
	// Align Begin line
    UpdateTop(TW) -= UpdateTop(TW)%(LINE_HEIGHT(TW));
    BeginRow = TW->TopRow + UpdateTop(TW)/(LINE_HEIGHT(TW));
    EndRow = TW->TopRow + 1 + UpdateBottom(TW)/(LINE_HEIGHT(TW));
	BeginCol = 0;
	EndCol = TW->Cols + 1;

#ifdef _DETECT_BUGS_
	if ( CheckBug( TW, 10005,
        (TW->TopRow != 0 && TWFileValid(TW))))
			return ;
#endif

	FontLead = FONT_LEADING(TW);

	if ( EndRow > TW->BufRows ) EndRow = TW->BufRows;
	switch( Encoding ) {
    case BIG5:  RefreshScreenBig5(TW); break;
    case HZ:    RefreshScreenHZ(TW);  break;
    case ASCII: RefreshScreenASCII(TW); break;
    default:    RefreshScreenGB(TW);  break;
	}
}

//
// Redisplay buffer to screen using GB encoding scheme
//
void PASCAL RefreshScreenGB( TextWin *TW )
{
	HDC hdc;
	PAINTSTRUCT ps;
	char *Buf;
	int RowIdx, ColIdx, FontIdx, PixRow, ASCLen = 0, GBIndex;
	char ChIdx;
	LPSTR ASCPtr;

	hdc = BeginPaint( TW->hWnd, &ps );

	SetBkMode(hdc, TRANSPARENT );

	SetTextAlign(hdc, TA_UPDATECP); // updates current position
	PixRow = LINE_SPACE(TW) + UpdateTop(TW) - HZLeading;
	FontLead -= HZLeading;
	Buf = TW->Buf + BufRowPtr(BeginRow);

    for( RowIdx = BeginRow; RowIdx < EndRow; RowIdx ++ ) {
        MoveTo( hdc, LEFT_MARGIN(TW), PixRow );
        for( ColIdx = BeginCol;
                ColIdx < EndCol && Buf[ColIdx] != 0; ColIdx++) {
			if ( Buf[ColIdx] > 0xa0 ) {
				char ChByte1, ChByte2;
				ChByte1 = Buf[ColIdx++];
				ChByte2 = Buf[ColIdx];
				if ( ASCLen ) {
					SelectObject( hdc, ASCIIFont );
					MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)),
						PixRow - FontLead );
					TextOut( hdc, 0, 0, ASCPtr, ASCLen );
					MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)), PixRow );
					ASCLen = 0;
				}
				GBIndex = ( ChByte1 - 0xa1)*94 + ( ChByte2 - 0xa1);
				if (ChByte1 >= 0xB0)	GBIndex -= (int)FontGAP;
				FontIdx = GBIndex / 224;
				ChIdx = GBIndex % 224 + 0x20;
				SelectObject (hdc, hChFont[FontIdx]);
				TextOut (hdc, 0, 0, (LPSTR) &ChIdx, 1);
			} else {
				if ( !ASCLen ) ASCPtr = (LPSTR)&Buf[ColIdx];
				ASCLen ++;
			}
		}
		if ( ASCLen ) {
			SelectObject( hdc, ASCIIFont );
			MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)),
				PixRow - FontLead );
			TextOut( hdc, 0, 0, ASCPtr, ASCLen );
			ASCLen = 0;
		}
        PixRow += LINE_HEIGHT(TW);
		Buf = NextRowPtr(Buf);
	}
    EndPaint( TW->hWnd, &ps );
} // End of RefreshScreenGB

void PASCAL RefreshScreenBig5( TextWin *TW )
{
	HDC hdc;
	PAINTSTRUCT ps;
	char *Buf;
	int RowIdx, ColIdx, FontIdx, PixRow, ASCLen = 0, B5Index = -1;
	char ChIdx;
	LPSTR ASCPtr;

	hdc = BeginPaint( TW->hWnd, &ps );

	SetBkMode(hdc, TRANSPARENT );

	SetTextAlign(hdc, TA_UPDATECP); // updates current position
    PixRow = LINE_SPACE(TW) + UpdateTop(TW) - HZLeading;
	FontLead -= HZLeading;
	Buf = TW->Buf + BufRowPtr(BeginRow);

    for( RowIdx = BeginRow; RowIdx < EndRow; RowIdx ++ ) {
		BYTE C1, C2;
        MoveTo( hdc, LEFT_MARGIN(TW), PixRow );
		for( ColIdx = BeginCol;
				ColIdx < EndCol && Buf[ColIdx] != 0; ColIdx ++ ) {
            C1 = (BYTE)Buf[ColIdx];
            C2 = (BYTE)Buf[ColIdx+1];
/* Big5 encodings
First byte:     0xA1 - 0xF9
        symbols:        0xA1 - 0xA3     ( 3 sectors)
        common hanzi:   0xA4 - 0xC6     (35 sectors)
        undefined:      0xC7 - 0xC8     ( 2 sectors)
        rare hanzi:     0xC9 - 0xF9     (49 sectors)
        total defined:  3+35+2          (87 sectors)
Second byte:    0x40 - 0xFE
        part one:       0x40 - 0x7E     (63 codes)
        undefined:      0x7F - 0xA0     (34 codes)
        part two:       0xA1 - 0xFE     (94 codes)
		total defined:  63+94          (157 codes)
coding space:   87 x 157 = 13659 codes
*/

			if ( (C1 > 0xa0 && C1 < 0xc7) || (C1 > 0xc8 && C1 < 0xfa) ) {
				if ( C2 > 0x3f && C2 <  0x7f )
					B5Index = ((C1 - 0xa1)*157 + (C2 - 0x40));
				else if ( C2 > 0xa0 && C2 < 0xff )
					B5Index = ((C1 - 0xa1)*157 + (C2 - 0xa1) + 63);
				if ( C1 > 0xc8 )
					B5Index -= FontGAP;
			}
			if ( B5Index < 0 ) {
				if ( !ASCLen ) ASCPtr = (LPSTR)&Buf[ColIdx];
				ASCLen ++;
			} else {
				ColIdx ++;
				if ( ASCLen ) {
					SelectObject( hdc, ASCIIFont );
					MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)),
						PixRow - FontLead );
					TextOut( hdc, 0, 0, ASCPtr, ASCLen );
					MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)), PixRow );
					ASCLen = 0;
				}
				FontIdx = B5Index / 224;
				ChIdx = B5Index % 224 + 0x20;
				B5Index = -1;
				SelectObject (hdc, hChFont[FontIdx]);
				TextOut (hdc, 0, 0, (LPSTR) &ChIdx, 1);
			}
		}
		if ( ASCLen ) {
			SelectObject( hdc, ASCIIFont );
			MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)),
				PixRow - FontLead );
			TextOut( hdc, 0, 0, ASCPtr, ASCLen );
			ASCLen = 0;
		}
        PixRow += LINE_HEIGHT(TW);
		Buf = NextRowPtr( Buf );
	}
    EndPaint( TW->hWnd, &ps );

} // End of RefreshScreenBig5

void PASCAL RefreshScreenHZ( TextWin *TW )
{
	HDC hdc;
	PAINTSTRUCT ps;
	BOOL HZOn;
	char *Buf, ChIdx;
	int RowIdx, ColIdx,  ColIdxScr, FontIdx, GBIndex, PixRow;
	char ASCIIStr[COLINC], ASCIdx = 0;

    hdc = BeginPaint( TW->hWnd, &ps );

	SetBkMode(hdc, TRANSPARENT );

	SetTextAlign(hdc, TA_UPDATECP); // updates current position
	PixRow = LINE_SPACE(TW) + UpdateTop(TW) - HZLeading;
	FontLead -= HZLeading;
	Buf = TW->Buf + BufRowPtr(BeginRow);

    for( RowIdx = BeginRow; RowIdx < EndRow; RowIdx ++ ) {
		MoveTo( hdc, LEFT_MARGIN(TW), PixRow );
		HZOn = FALSE;
		for( ColIdx = ColIdxScr = BeginCol; Buf[ColIdx] != 0 &&
				ColIdxScr < EndCol; ColIdx ++ ) {
			if ( Buf[ColIdx] == '~' ) {
                ColIdx++;
				if ( HZOn )	{ // in Chinese mode.
                    if ( Buf[ColIdx] == '}' ) {
						HZOn = 0;
						if ( ShowGBHZMark ) {
							ASCIIStr[ASCIdx++] = '~';
							ASCIIStr[ASCIdx++] = '}';
						}
						continue;
					} 
                }
				if ( Buf[ColIdx] == '{'  ) { // Chinese
					HZOn = 1;
					if ( ShowGBHZMark ) {
						ASCIIStr[ASCIdx++] = '~';
						ASCIIStr[ASCIdx++] = '{';
					}
					if ( ASCIdx ) {
						ColIdxScr += ASCIdx;
						SelectObject( hdc, ASCIIFont );
						MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)),
							PixRow - FontLead );
						TextOut( hdc, 0, 0, ASCIIStr, ASCIdx );
						MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)), PixRow );
						ASCIdx = 0;
					}
				} else if ( Buf[ColIdx] == '~' ) {
					ASCIIStr[ASCIdx++] = '~';
				} else {
					if ( Buf[ColIdx] == 0 && ShowGBHZMark )
						ASCIIStr[ASCIdx++] = '~';
					ColIdx --;
				}
			} else if ( HZOn ) {	// In Chinese mode.
                char ChByte1, ChByte2;
				ChByte1 = Buf[ColIdx++];
				ChByte2 = Buf[ColIdx];
				GBIndex = ( ChByte1 - 0x21 )*94 + ( ChByte2 - 0x21 );
				if (ChByte1 >= 0xB0-0xa1+0x20)
					GBIndex -= (int)FontGAP;
				FontIdx = GBIndex / 224;
				ChIdx = GBIndex % 224 + 0x20;
       			SelectObject (hdc, hChFont[FontIdx]);
				TextOut (hdc, 0, 0, (LPSTR) &ChIdx, 1);
			} else {
                ASCIIStr[ASCIdx++] = Buf[ColIdx];
			}
		}
		if ( ASCIdx ) {
			SelectObject( hdc, ASCIIFont );
			MoveTo( hdc, LOWORD(GetCurrentPosition(hdc)),
				PixRow - FontLead );
			TextOut( hdc, 0, 0, ASCIIStr, ASCIdx );
			ASCIdx = 0;
		}
        PixRow += LINE_HEIGHT(TW);
		Buf = NextRowPtr(Buf);
	}
    EndPaint( TW->hWnd, &ps );

} // End of RefreshScreenHZ


void PASCAL RefreshScreenASCII( TextWin *TW )
{
	HDC hdc;
	PAINTSTRUCT ps;
	char *Buf;
	int RowIdx, ColIdx, PixRow;

	hdc = BeginPaint( TW->hWnd, &ps );
	SelectObject( hdc, ASCIIFont );
	SetTextAlign(hdc, TA_UPDATECP); // updates current position

    // Only Chinese fonts need a lead
	PixRow = LINE_SPACE(TW) + UpdateTop(TW) - FontLead;
	Buf = TW->Buf + BufRowPtr(BeginRow);
	for( RowIdx = BeginRow; PixRow < (int)TW->Height && RowIdx < EndRow;
            PixRow += LINE_HEIGHT(TW), RowIdx ++ ) {
		MoveTo( hdc, LEFT_MARGIN(TW), PixRow );
		TextOut( hdc, 0, 0, Buf, lstrlen(Buf) );
		Buf = NextRowPtr(Buf);
	}
    EndPaint( TW->hWnd, &ps );

} // End of RefreshScreenASCII


BOOL PASCAL ChangeEncode( TextWin *TW, int NewEncode )
{
    int OCHeight = CHAR_HEIGHT(TW);

	HZOnMark[0] = HZOffMark[0] = 0;
	HzOff = 0;
	KXGetChnCompReset();
	switch( NewEncode ) {
	case ASCII:	Encoding = NewEncode; break;
	case BIG5:
        if ( !CurrentFont || CurrentFont->Encoding != BIG5 ) {
            InitChFont(TW, &B5Font );
		}
		Encoding = BIG5;
        break;
	case HZ:
		HZOnMark[0] = HZOffMark[0] = '~';
		HzOff = 2;
	case GB:
        if ( !CurrentFont || CurrentFont->Encoding != GB ) {
            InitChFont(TW, &GBFont);
		}
		Encoding = NewEncode;
		break;
	default:
        return FALSE;
	}
	if ( OCHeight != CHAR_HEIGHT(TW) ) {
		GetWindowSize();
		FileDisp.ReloadBuffer = 1;
		AdjustTextWin( &FileDisp, TRUE );
		AdjustTextWin( &KXDisp, TRUE );
	}
	if ( !FileDisp.AssocFile->FilePtr )
		GetBanner(&FileDisp); // Change banner encoding
	CheckEncodeMenu();
	CheckZiTiMenu();
	CheckSizeMenu();
	InvalidateRect( ThisApphWnd, NULL, TRUE );
	return TRUE;
}


