#define STRICT      // Strict type checking
#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include "kxmenuid.h"	// Menu identifiers
#include "kxglobal.h"   // Global variables
#include "kxwproto.h"	// Function Prototypes

BOOL PASCAL EraseBuffer( TextWin *TW, int r1, int r2 )
{
	char *Buf;
	RECT RC;
	int I;

	if ( r1 > r2 || r1 < 0 || r2 >= TW->BufRows ) return FALSE;
	I = r1;
	Buf = TW->Buf + BufRowPtr(r1);
	while( I++ <= r2 ) {
		Buf[0] = 0;
        Buf = NextRowPtr(Buf);
	}
	if ( r2+1 == TW->BufRows ) {
		TW->BufRows = r1+1;
		AdjustTextWin( TW, FALSE );
	}
	GetClientRect( TW->hWnd, &RC );
	RC.top = (r1 - TW->TopRow ) * LINE_HEIGHT(TW);
	RC.bottom = ((min(TW->Rows, r2)+1) - TW->TopRow) * LINE_HEIGHT(TW);

	InvalidateRect( TW->hWnd, &RC, TRUE );
	UpdateWindow( TW->hWnd );

	return TRUE;
}

int PASCAL PrintToBuffer( TextWin *TW, int r, int c, char *s, int wrap )
{
	RECT RC;
	char *Buf;
	int I, J = 0, R, WrapCol, NextRow = 0;

    if ( r < 0 || r >= TW->MaxRows || c < 0 || c >= MAXCOL )
		return 0;
	if ( r >= TW->BufRows ) {
		// make all the non valid rows valid and blank
        I = TW->BufRows;
        Buf = TW->Buf + BufRowPtr(I);
		while( I++ < r ) {
			Buf[0] = 0;
            Buf = NextRowPtr(Buf);
		}
	} else
        Buf = TW->Buf + BufRowPtr(r);
	while( I < MAXCOL && Buf[I] != 0 ) I++;
	if ( I >= c )
		I = c;
	else {
		I --;
		while( I < c ) Buf[I++] = ' ';
	}

	R = r;
	if ( wrap == TW_WRAPWIN ) WrapCol = min(TW->Cols-1, MAXCOL);
	else 					  WrapCol = MAXCOL;
	while( J < lstrlen((LPCSTR) s) ) {
		if ( I >= WrapCol ) {
			if ( wrap == TW_WRAPNON ) break;
			NextRow = 1;
		}
		if ( s[J] == '\n' ) { J ++; NextRow = 1; }
		if ( NextRow ) {
			NextRow = 0;
			R ++;
			if ( R < TW->MaxRows ) {
				Buf[I] = 0;
				I = 0;
				Buf = NextRowPtr( Buf );
			} else {
				R --;
				break;
            }
        }
		Buf[I++] = s[J++];
	}
	R ++;
	if ( R > TW->BufRows ) {
		TW->BufRows = R;
		if ( TW->BufRows > TW->Rows ) {
			AdjustTextWin( TW, FALSE );
		} else
			SetScrollRange( TW->hWnd, SB_VERT, 0, TW->BufRows, TRUE );
	}
	Buf[I] = 0;
    GetClientRect( TW->hWnd, &RC );
    RC.top = (r - TW->TopRow) * LINE_HEIGHT(TW);
    RC.bottom = (R - TW->TopRow) * (LINE_HEIGHT(TW));
	InvalidateRect( TW->hWnd, &RC, TRUE );
	UpdateWindow( TW->hWnd );

	return R-r;
}

//
// Load enough lines from the opened file pointer FilePtr into ScrBuf.
//
BOOL PASCAL LoadToScrBuffer( TextWin *TW )
{
	FILE *FilePtr;
	char *Buf = TW->Buf;
    int i, j, r = LINE_SPACE(TW), Len;

	TW->ReloadBuffer = 0; // We are about to load!
	if ( TW->AssocFile ) {
		FilePtr = TW->AssocFile->FilePtr;
		if ( FilePtr == NULL ) return FALSE;
	} else
        return FALSE;

    fseek( FilePtr, TW->AssocFile->FileMark, SEEK_SET );
    TW->BufRows = 0;  // Reset lines in the buffer

    for( i = 0; r < (int)TW->Height && i < TW->MaxRows; i ++ ) {
        if ( fgets( CommonBuffer, CBUFSIZE, FilePtr ) == NULL ) break;
		Len = lstrlen( (LPCSTR)CommonBuffer );
		if ( Len > MAXCOL ) Len = MAXCOL;
		lstrcpyn( (LPSTR)Buf, (LPCSTR)CommonBuffer, Len+1 );
        if ( Len > 1 && Buf[Len-2] <= 13 ) Buf[Len-2] = 0;
        else if ( Buf[Len-1] <= 13 ) Buf[Len-1] = 0;
        r += LINE_HEIGHT(TW);
        Buf = NextRowPtr(Buf);
        TW->BufRows ++;
	}
    if ( UpdateBottom(TW) > r )  UpdateBottom(TW) = r;
	return TRUE;
} // End of LoadToScrBuffer

//
// Open the file (szFileName) for reading
//
BOOL PASCAL ReadFile( TextWin *TW )
{
	char *s;
	FileInfo *FI;
    int i, LongLines = 0;

    FI = TW->AssocFile;
    if ( FI )
        s = FI->szFileName;
    else
        return FALSE;

	// Check encoding scheme by the file extension
	while( *s != 0 ) s ++;
    while( s != FI->szFileName ) {
		if ( *s == '.' ) {
			s ++;
			switch( toupper(*s) ) {
			case 'H':	ChangeEncode( TW, HZ ); break;
			case 'B':	ChangeEncode( TW, BIG5 ); break;
			case 'G':	ChangeEncode( TW, GB ); break;
			default:	ChangeEncode( TW, ASCII ); break;
			}
			break;
		}
		s --;
	}
	FI->FilePtr = fopen( FI->szFileName, "rb" );
    if ( FI->FilePtr == NULL ) {
		wsprintf( CommonBuffer, "Error: Can't open file %s.",
                (LPSTR)FI->szFileName );
		MessageBox( ThisApphWnd, CommonBuffer, "Open File",
				MB_OK|MB_APPLMODAL|MB_ICONEXCLAMATION );
        return FALSE;
	}
    PrintToShengZi( "\n**** New File: ", "" );
	PrintToShengZi( FI->szFileName, "\n" );

    wsprintf( CommonBuffer, "KXWin: %s", (LPSTR)FI->szFileName );
	SetWindowText( ThisApphWnd, CommonBuffer );

	/* Initialize LineSize array */
	for( i = 0; ; i++ ) {
    	int Len;
        if ( fgets( CommonBuffer, CBUFSIZE, FI->FilePtr ) == NULL ) break;
        if ( i < MAXLINES ) {
			Len = lstrlen( (LPCSTR)CommonBuffer );
			if ( Len > MAXCOL ) LongLines ++;
			if ( Len > 0xff ) {
				wsprintf( CommonBuffer,
					"ERROR!!!  This program definitely cannot handle lines \n\
longer than 256 chars but line %d is %d chars long.", i+1, Len );
				MessageBox( ThisApphWnd, CommonBuffer, "Fatal Error",
					MB_APPLMODAL|MB_ICONSTOP|MB_OK );
				CloseFile(TW);
				return FALSE;
            }
			FI->LineSize[i] = (char)Len;
		}
	}
	if ( i == 0 ) {
		wsprintf( CommonBuffer, "File \"%s\" is empry (zero bytes).",
			FI->szFileName );
		MessageBox( ThisApphWnd, CommonBuffer, "Read File", 
				MB_OK|MB_APPLMODAL|MB_ICONEXCLAMATION );
        CloseFile(TW);
		return FALSE;
	}
    FI->TotalLines = i;
    if ( FI->TotalLines >= MAXLINES ) {
		wsprintf( CommonBuffer,
			"Sorry, this file has too many lines (%d).\n\
I can only handle files with up to %d lines.", FI->TotalLines, MAXLINES );
		MessageBox( ThisApphWnd, CommonBuffer, "Read File",
				MB_OK|MB_APPLMODAL|MB_ICONEXCLAMATION );
        CloseFile(TW);
        GetBanner(TW);
        return FALSE;
	}

	if ( LongLines ) {
		wsprintf( CommonBuffer,
			"%d lines are longer than can be handled.\n\
Maximum length is %d ASCII characters.", LongLines, MAXCOL );
		MessageBox( ThisApphWnd, CommonBuffer, "Read File",
				MB_OK|MB_APPLMODAL|MB_ICONEXCLAMATION );
	}

	/* Load the file into the screen buffer */
	rewind( FI->FilePtr );
	TW->ScrollAble = FALSE;
    AdjustTextWin(TW, TRUE);
	if ( LoadToScrBuffer(TW) ) {
		SetScrollRange( TW->hWnd, SB_VERT, 0,
        	TW->AssocFile->TotalLines, FALSE );
        SetScrollPos( TW->hWnd, SB_VERT, 0, TRUE );
        InvalidateRect( TW->hWnd, NULL, TRUE );
		return TRUE;
	}
    return FALSE;
} // End of ReadFile

//
// Ask user to choose a new file name and open it for reading
//
BOOL PASCAL OpenNewFile( TextWin *TW )
{
	OPENFILENAME ofn;
	UINT  i, cbString;
	char  chReplace;    /* string separator for szFilter */
    char  FileName[256];
	char  szFilter[256];
	char  szFileTitle[256];
	static char szDirName[256];
	static BOOL Init = FALSE;
	int FilterIdx;
    FileInfo *FI = TW->AssocFile;

	if ( !FI ) return FALSE;

	if ( !Init ) {
		szDirName[0] = '\0';
		FI->szFileName[0] = '\0';
		Init = TRUE;
	}
	switch( Encoding ) {
	case BIG5:  FilterIdx = 1; break;
	case GB:    FilterIdx = 2; break;
	case HZ:    FilterIdx = 3; break;
	default:    FilterIdx = 4; break;
    }

	if ((cbString = LoadString(ThisApphInst, IDS_FILTERSTRING,
       		szFilter, sizeof(szFilter))) == 0) {
		// Handles error here.
		return FALSE;
	}
	chReplace = szFilter[cbString - 1]; /* retrieve wildcard */

	for (i = 0; szFilter[i] != '\0'; i++) {
    	if (szFilter[i] == chReplace)
    		szFilter[i] = '\0';
	}

	/* Set all structure members to zero. */
	memset(&ofn, 0, sizeof(OPENFILENAME));

	lstrcpy( FileName, FI->szFileName );
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = ThisApphWnd;
	ofn.lpstrFilter = szFilter;
	ofn.nFilterIndex = FilterIdx;
    ofn.lpstrFile= FileName;
    ofn.nMaxFile = sizeof(CommonBuffer);
	ofn.lpstrFileTitle = szFileTitle;
	ofn.nMaxFileTitle = sizeof(szFileTitle);
	ofn.lpstrInitialDir = szDirName;
	ofn.Flags = OFN_SHOWHELP | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

	if (GetOpenFileName(&ofn)) {
        BOOL R;

		FilterIdx = ofn.nFilterIndex;
		CloseFile(TW);
        lstrcpy( FI->szFileName, FileName );
        R = ReadFile(TW);
        if ( R ) {
			FI->BookMarkLine = GetPrivateProfileInt( "KXBookMark",
									(LPCSTR)FileName, -1, "kxwin.ini" );
		}
        return R;
	}
	return FALSE;
/*
	else {
		// Handles errors here
	}
*/
}  // End of OpenNewFile

void PASCAL SaveBookMark( FileInfo *FI, int force )
{
	if ( FI && FI->BookMarkLine != FI->FileLine ) {
		int Ans;

        if ( ! FI->FilePtr ) {
        		MessageBox( ThisApphWnd, "Error: No document is open.",
		        	"Save Bookmark", MB_OK|MB_ICONEXCLAMATION );
                return;

        }
		wsprintf( CommonBuffer, "Bookmark is different from current \
possition:\n    Bookmark is line:\t%d\n    Current mark is line:\t%d\n\
Would you like to save the Bookmark?",
			FI->BookMarkLine,  FI->FileLine );
		Ans = MessageBox( ThisApphWnd, CommonBuffer, "Save Bookmark",
					MB_APPLMODAL|MB_YESNO|MB_ICONQUESTION );
		if ( Ans == IDYES ) {
        	FI->BookMarkLine = FI->FileLine;
			wsprintf( CommonBuffer, "%d", FI->BookMarkLine );
			Ans = WritePrivateProfileString( "KXBookMark",
				(LPCSTR)FI->szFileName, (LPCSTR)CommonBuffer, "kxwin.ini" );
			if ( Ans == 0 ) {
				MessageBox( ThisApphWnd,
					"Saving bookmark to kxwin.ini failed.", "Error!",
					MB_OK );
            }
		}
	} else if ( force )
		MessageBox( ThisApphWnd, "This position is already saved to bookmark.",
			"Save Bookmark", MB_OK|MB_ICONEXCLAMATION );
}

void PASCAL GotoBookMark( FileInfo *FI )
{
	if ( FI && FI->BookMarkLine >= 0 &&
			FI->BookMarkLine != FI->FileLine ) {
		FileVertScroll( FI->BookMarkLine - FI->FileLine );
	} else
		MessageBox( ThisApphWnd,
			"Either you are already at the bookmark\n\
or you haven't save any bookmark yet.",
		"Save Bookmark", MB_OK|MB_ICONEXCLAMATION );
}

void PASCAL CloseFile( TextWin *TW )
{
    FileInfo *FI = TW->AssocFile;
    TW->ScrollAble = FALSE;
	if ( !FI ) return;

    if ( FI->FilePtr ) {
	// Check if need to save bookmark, don't force it though.
		SaveBookMark( FI, 0 );  

	/* Now close old file */
		fclose(FI->FilePtr);

		PrintToShengZi( "\n**** Close file: ", "" );
        PrintToShengZi( FI->szFileName, "\n" );
		CloseShengZiFile();
		SetWindowText( ThisApphWnd, "KXWin" );
	}
    FI->FilePtr = 0;
    FI->FileMark = 0;
	FI->FileLine = 0;
	FI->BookMarkLine = -1;
}

struct myText{
    char *s;
};

void PASCAL GetBanner( TextWin *TW )
{
    int BufI = 0, CI;
    char *Buf = TW->Buf;
    static struct myText Eng[] = {
		"KuaiXue Chinese Teaching Tool for Windows version 1.0",
		"Written by Joel Chen under the supervision of",
		"",
        "Dr. Norman Matloff",
        "Computer Science Department ",
        "University of California, Davis ", 0 };
	static struct myText Chn[] = {
		// Big5
        " ",
		"־ǤоǤu",
        "@: qE",
        "[{jǹqt",
        0,
        // HZ
        " ",
		"~{?lQ'VPND=LQ'9$>_~}",
        "~{WwU_#:3BPc:F~}",
        "~{<SV]4sQ'5gDT?FO5~}",
        0,
        // GB
		" ",
		"ѧĽѧ",
        ": ",
        "ݴѧԿϵ", 0 };
    for( BufI = 0; Eng[BufI].s; BufI++ ) {
        wsprintf( Buf, Eng[BufI].s );
        Buf = NextRowPtr(Buf);
    }
    switch( Encoding ) {
    case BIG5:  CI = 0*5;   break;
    case HZ:    CI = 1*5;   break;
    case GB:    CI = 2*5;   break;
    default:    CI = 3*5-1; break;
    }
	for( ; Chn[CI].s; CI++, BufI++) {
        wsprintf( Buf, Chn[CI].s);
        Buf = NextRowPtr(Buf);
    }
	TW->TopRow = 0;
	TW->BufRows = BufI;
    if ( TW->hWnd ) {
		SetScrollRange( TW->hWnd, SB_VERT, 0, BufI, FALSE );
		AdjustTextWin( TW, FALSE );
	}
}


int PASCAL SeekForward( TextWin *TW, int Lines )
{
	int i, LinesLeft;
    FileInfo *FI = TW->AssocFile;
    long SeekBytes = FI->FileMark;

    LinesLeft = FI->TotalLines - FI->FileLine - 1;
    if ( LinesLeft < TW->ScrollRowLimit )
		return 0;

	LinesLeft -= Lines;

    if ( LinesLeft < TW->ScrollRowLimit ) {  // End of file reached
        Lines = FI->TotalLines - FI->FileLine - TW->ScrollRowLimit;
		if ( Lines < 1 ) return 0;
	}

    for( i = FI->FileLine; i < FI->TotalLines && i < FI->FileLine+Lines; i ++ )
        SeekBytes += FI->LineSize[i];

	FI->FileMark = SeekBytes;
    FI->FileLine += Lines;
	if ( FI->FileLine > FI->TotalLines )
		FI->FileLine = FI->TotalLines - 1;

	return Lines;
} // End of SeekForward


int PASCAL SeekBackward( TextWin *TW, int Lines )
{
	int i;
    FileInfo *FI = TW->AssocFile;
    long SeekBytes = FI->FileMark;

    if ( FI->FileLine == 0 ) return 0;
    if ( FI->FileLine - Lines < 0 ) Lines = FI->FileLine;
    for( i = FI->FileLine-1; i >= 0 && i > FI->FileLine - Lines - 1; i -- )
        SeekBytes -= FI->LineSize[i];

	FI->FileMark = SeekBytes;
    FI->FileLine -= Lines;
	return Lines;
}

