#define STRICT      // Strict type checking
#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include "kxmenuid.h"	// Menu identifiers
#include "kxglobal.h"   // Global variables
#include "kxwproto.h"	// Function Prototypes


char* GetToken(char *S, char T, int n, int *TSize, int TrimSpace )
{
    int Tn = 0, size = 0, I;
    char *TStart;
    if ( TrimSpace ) {
        if ( T == ' ' ) return 0;
        while( *S && *S == ' ' ) S++;
    }
    TStart = S;
    while( 1 ) {
        if (*S == T || *S == 0) {
            Tn ++;
            if ( Tn == n ) {
                S --;
                if ( TrimSpace ) {
                    while( size > 0 && *S != T && *S == ' ') {
                        S --;
                        size --;
                    }
                }
				if ( TSize ) *TSize = size;
                return TStart;
            } else if ( *S != 0 ) {
                S ++;
                size = 0;
                if ( TrimSpace ) {
                    while( *S && *S == ' ' ) S++;
                }
                TStart = S;
                continue;
            }
            return 0;
        }
        size ++;
        S ++;
    }
}


BOOL PASCAL ReadFontInfoFromIni( ChFontInfo *FInfo )
{
    char Entry[30];
	int TkSize;
	char *TkPtr;

	if ( FInfo->Encoding == ASCII ) return FALSE;
    wsprintf( Entry, "%s %dx%d %s Font",
        (LPSTR)EncodeStr[FInfo->Encoding], FInfo->Size, FInfo->Size,
		(LPSTR)(FInfo->JianTi ? "JianTi" : "FanTi") );

	while ( GetPrivateProfileString( "KuaiXue", Entry, 0,
            CommonBuffer, CBUFSIZE, "kxwin.ini" ) ) {
		TkPtr = GetToken( CommonBuffer, ',', 1, &TkSize, 1 );
        if ( TkPtr && TkSize < 13 ) {
            strncpy( FInfo->FileName, TkPtr, TkSize );
            FInfo->FileName[TkSize] = 0;
        } else
            break;
		TkPtr = GetToken( CommonBuffer, ',', 2, &TkSize, 1 );
        if ( TkPtr && TkSize < 33 ) {
            strncpy( FInfo->FontName, TkPtr, TkSize );
            FInfo->FontName[TkSize] = 0;
        } else
			break;
		TkPtr = GetToken( CommonBuffer, ',', 3, &TkSize, 1 );
        if ( TkPtr ) {
			FInfo->Gap = atoi(TkPtr);
        } else
			break;
		TkPtr = GetToken( CommonBuffer, ',', 4, &TkSize, 1 );
		if ( TkPtr ) FInfo->MaxFonts = atoi(TkPtr);
		return TRUE;
	}
	wsprintf( CommonBuffer, "No information for the Chinese font you \
requested\ncan be found in kxwin.ini which is in your windows\n\
directory.  Please double check that the entry:\n\n%s\n\nis \
actually refering to something that exists.\n\nAlso make sure that all \
other entries are correct.\nPlease read the file kxwin.txt if \
you need help.\n\nAs a result of this error , the font \n\
will either not change or the \ndefault will be used.",
		(LPSTR)Entry );
	MessageBox( NULL, CommonBuffer, "KXWin Load Font Info.",
				MB_OK|MB_APPLMODAL|MB_ICONSTOP );
	return FALSE;
}


BOOL PASCAL InitAppMem( void )
{
	// Get memory for file. This only keep the size of each
	// line in the file for seeking.  Windows file cache should
	// do a pretty good job so real disk access is minimized.
	FileDisp.AssocFile->hLineSize = LocalAlloc( LMEM_FIXED|LMEM_ZEROINIT, MAXLINES );
	if ( !FileDisp.AssocFile->hLineSize ) {
		MessageBox( ThisApphWnd, "LocalAlloc for memory failed.",
					"KXWin", MB_OK );
		return FALSE;
	}
	FileDisp.AssocFile->LineSize = (char*)LocalLock( FileDisp.AssocFile->hLineSize );


	DictMem = GlobalAlloc( GMEM_FIXED, MaxDictSize );
	if ( !DictMem ) {
		MessageBox( ThisApphWnd, "GlobalAlloc for memory failed.",
					"KXWin", MB_OK );
		return FALSE;
	}
    Dict = (char FAR *)GlobalLock( DictMem );
	CompDictMem = GlobalAlloc( GMEM_FIXED, MaxDictSize );
	if ( !CompDictMem ) {
		MessageBox( ThisApphWnd, "GlobalAlloc for memory failed.",
					"KXWin", MB_OK );
		return FALSE;
	}
	CompDict = (char FAR *)GlobalLock( CompDictMem );

	return TRUE;
}

BOOL PASCAL InitTextWin( TextWin *TW,
                  HINSTANCE hInst,
                  HWND ParenthWnd )
{
    TW->hWnd = CreateWindow(
                     "KXTextWin",
					 "",
					 WS_CHILD|WS_VSCROLL,
					 0, 0, 50, 50,
                     ParenthWnd,
                     NULL,
                     hInst,
					 NULL );

    TW->ReloadBuffer = FALSE;
    TW->ScrollAble = FALSE;
    if ( TW->hWnd ) return TRUE;
	return FALSE;
}


void LoadValuesFromIni( void )
{
	char *TkPtr;
	ChFontInfo *FtI = &GBFont;
    BOOL isJianTi = TRUE;

	if ( GetPrivateProfileString( "KuaiXue", "Mode", "GB", CommonBuffer,
				CBUFSIZE, "kxwin.ini" ) ) {
		TkPtr = GetToken( CommonBuffer, ',', 1, 0, 1 );
		if ( TkPtr )  
			switch( TkPtr[0] ) {
			case 'H':
			case 'h':
				Encoding = HZ;
                break;
			case 'g':
			case 'G':
				Encoding = GB;
				break;
			case 'B':
			case 'b':
				isJianTi = FALSE;
				Encoding = BIG5;
				FtI = &B5Font;
				break;
			default:
				return;
            }
	}
	FtI->JianTi = (BOOL)GetPrivateProfileInt( "KuaiXue", "Jian Ti Zi",
					(int)isJianTi, "kxwin.ini" );
	FtI->Size = (int)GetPrivateProfileInt( "KuaiXue", "Font Size",
					16, "kxwin.ini" );
	if ( FtI->Size != 16 && FtI->Size != 24 ) {
		MessageBox( NULL, "Warning:\nFont size is being set to 16 because \
the\nvalue in kxwin.ini is not 16 nor 24.", "KXWin Initialization",
		MB_OK|MB_ICONEXCLAMATION );
		FtI->Size = 16;
	}
	TrackScroll = !(BOOL)GetPrivateProfileInt( "KuaiXue", "Track Scroll",
					TRUE, "kxwin.ini" );
	LINE_SPACE((&FileDisp)) = (int)GetPrivateProfileInt( "KuaiXue",
					"Line Space", 0, "kxwin.ini" );
	ShowGBHZMark = !(BOOL)GetPrivateProfileInt( "KuaiXue",
					"Show HZ Escape Code", FALSE, "kxwin.ini" );
}


BOOL PASCAL EnterApp( HINSTANCE hInstance, // which program are we?
					  HINSTANCE hPrevInst, // is there another one?
					  LPSTR lpszCmdLine,   // Command line arguments
                      int nCmdShow )       // Window size (icon, etc)
{
	long I;
	ChFontInfo FtI1, FtI2;

	InitGlobal();

	memcpy( &FtI1, &GBFont, sizeof( ChFontInfo ));
	memcpy( &FtI2, &B5Font, sizeof( ChFontInfo ));
	LoadValuesFromIni();
	if ( !ReadFontInfoFromIni( &GBFont ) )
		memcpy( &GBFont, &FtI1, sizeof( ChFontInfo ));
	if ( !ReadFontInfoFromIni( &B5Font ))
		memcpy( &B5Font, &FtI2, sizeof( ChFontInfo ));

	// Initialize the application and an instance
	if ( !InitKXWindow( hInstance, hPrevInst )) {
		// Error Handling
        return FALSE;
	}

	ShowWindow( ThisApphWnd, nCmdShow );    // Make main window Visible
	if ( !InitAppMem() ) return FALSE;

    // Initialize File display text area struct

	InitTextWin( &FileDisp, hInstance, ThisApphWnd );
	InitTextWin( &KXDisp, hInstance, ThisApphWnd );
	ShowWindow( FileDisp.hWnd, nCmdShow );
	ShowWindow( KXDisp.hWnd, nCmdShow );

    // Initialize Chinese Encoding & Font

	ChangeEncode( &FileDisp, Encoding );
	CheckEncodeMenu();                      // Check encoding menu item
    CheckGBHZMarkMenu();                    // Check HZ Show mark menu item
	CheckTScrollMenu();                     // Check track scroll menu item
	CheckZiTiMenu();
	CheckSizeMenu();

	if ( lstrlen(lpszCmdLine) > 0 ) {
		lstrcpy( FileDisp.AssocFile->szFileName, lpszCmdLine );
		ReadFile(&FileDisp);
	}
//	MessageBox( ThisApphWnd, "Go", "Go", MB_OK );
    return TRUE;
}

//
// Register the application Class and create a window
//

BOOL PASCAL InitKXApp( HINSTANCE hInst )
{
	WNDCLASS wndclass;   // Window class structure

	wndclass.style          = CS_HREDRAW|CS_VREDRAW; // default style
    wndclass.lpfnWndProc    = KXWndProc;        // WndProc address
    wndclass.cbClsExtra     = 0;                // no extra class data
    wndclass.cbWndExtra     = 0;                // no extra window data
	wndclass.hInstance      = hInst;        	// which program?
                                                // stock arrow cursor
    wndclass.hCursor        = LoadCursor( NULL, IDC_ARROW );
                                                // stock blank cursor
	wndclass.hIcon          = LoadIcon( hInst, "KXLogoIcon");
    wndclass.lpszMenuName   = "KXMenu";
                                                // White background
    wndclass.hbrBackground  = GetStockObject( WHITE_BRUSH );
    wndclass.lpszClassName  = "KXWinClass";     // Window class name

    if ( RegisterClass(&wndclass) ) { // register the class
        wndclass.lpfnWndProc = KXTextWndProc;
        wndclass.lpszMenuName =  NULL;
        wndclass.lpszClassName = "KXTextWin";
        if ( RegisterClass(&wndclass) ) return TRUE;
    }
    return FALSE;
}

BOOL PASCAL InitKXWindow( HINSTANCE hInstance,
                   HINSTANCE hPrevInst )
{
	if ( !hPrevInst )
        if ( !InitKXApp(hInstance)) {
            // Error handling
            return FALSE;
        }
	ThisApphWnd = CreateWindow(
						 "KXWinClass",
                         "KXWin",
						 WS_OVERLAPPEDWINDOW,
                         CW_USEDEFAULT,
                         CW_USEDEFAULT,
						 CW_USEDEFAULT,
                         CW_USEDEFAULT,
                         NULL,
                         NULL,
                         hInstance,
						 NULL );

	ThisApphInst = hInstance;
	return TRUE;

}  // End of InitKXWindow


//
// Load the Chinese fonts and register them, and then
// create handles for all of them.
//
BOOL PASCAL InitChFont( TextWin *TW, ChFontInfo *NewFont )
{
	HDC hdc;
    HFONT oldhFont;
	TEXTMETRIC tm;
    int FontsAdded = 0;
	int i, FontHeight = 0, ChHeight;
	char *FontName;

	FontName = CommonBuffer +
		GetPrivateProfileString("KuaiXue", "FontDir", 0, CommonBuffer,
			CBUFSIZE, "kxwin.ini");

	if ( FontName == CommonBuffer ) {
		MessageBox( ThisApphWnd, "No Font dir specified.", "KXWin", MB_OK);
		DestroyWindow( ThisApphWnd );
		return FALSE;
	}

    FontName --;
	if ( *FontName != '\\' ) *FontName = '\\';
	FontName ++;

    if ( CurrentFont ) { // Deinit fonts
		DeleteChFont();
		wsprintf( FontName, (LPSTR)CurrentFont->FileName );
		RemoveFontResource(CommonBuffer);
		SendMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
	}

	if ( !NewFont ) return TRUE;  // Make sure we need to load a font

	wsprintf( FontName, "%s", (LPSTR)NewFont->FileName );
	FontsAdded = AddFontResource(CommonBuffer);

    if ( FontsAdded < 1 ) {
	   MessageBox( ThisApphWnd, "Load Font failed.", "KXWin", MB_OK);
	   DestroyWindow( ThisApphWnd );
	   return FALSE;
	}

	CurrentFont = NewFont;

	SendMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);

	if ( FontsAdded > 1 ) 	MaxFonts = FontsAdded;
	else 	MaxFonts = NewFont->MaxFonts;

	Encoding = NewFont->Encoding;
	FontGAP = NewFont->Gap;

	for( i = 0; i < sizeof(hChFont)/sizeof(hChFont[0]); i ++ ) {
		if ( i >= MaxFonts ) break;
		wsprintf( CommonBuffer, "%s_%02d", (LPSTR)NewFont->FontName, i+1 );
        hChFont[i] = CreateFont(CHAR_HEIGHT(TW), CCHAR_WIDTH(TW),
			0, 0, 0, 0, 0, 0, 1, 0, 0,
			DRAFT_QUALITY, 0, (LPCSTR)CommonBuffer);
		if ( hChFont[i] == NULL ) {
			MaxFonts = i;
			break;
		}
	}

    FontHeight = CHAR_HEIGHT(TW)/2;
	hdc = GetDC( ThisApphWnd );

    /* Get the size of the Chinese characters */
	oldhFont = SelectObject( hdc, hChFont[1] );
    GetTextMetrics( hdc, &tm );
	CCHAR_WIDTH(TW) = tm.tmAveCharWidth;
	CHAR_HEIGHT(TW) = tm.tmHeight;
	CHAR_WIDTH(TW) = CCHAR_WIDTH(TW)/2;
	SelectObject( hdc, oldhFont );

#define _simple_font_ 0

#if _simple_font_
 	oldhFont = SelectObject( hdc, ASCIIFont );
    GetTextMetrics( hdc, &tm );
    ASCIIFont = CreateFont( CHAR_HEIGHT(TW), CHAR_WIDTH(TW),
			0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, "Courier New" );
    FONT_LEADING(TW) = 0;
    LINE_SPACE(TW) = 2;
	LINE_HEIGHT(TW) = CHAR_HEIGHT(TW)+LINE_SPACE(TW);
#else
  while(1) {
        ASCIIFont = CreateFont( -FontHeight, CHAR_WIDTH(TW),
				0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, "Courier New" );
		oldhFont = SelectObject( hdc, ASCIIFont );
		GetTextMetrics( hdc, &tm );
		ChHeight = tm.tmHeight-3;
		if ( ChHeight >= CHAR_HEIGHT(TW)  )
			break;
		else
			DeleteObject(ASCIIFont);
		if ( ChHeight < CHAR_HEIGHT(TW) )
            FontHeight++;
		else if ( FontHeight > 0 )
            FontHeight--;
		else {
			// Panic!
		}
    }
	FONT_LEADING(TW) = tm.tmInternalLeading;
	if ( LINE_SPACE(TW) < 1 ) {
		LINE_SPACE(TW) = 0;
		if ( FONT_LEADING(TW) ) FONT_LEADING(TW) --;
		HZLeading = 0;
	} else 
		HZLeading = 1;

	LINE_HEIGHT(TW) = CHAR_HEIGHT(TW) + LINE_SPACE(TW) + 2;
#endif
	ReleaseDC( ThisApphWnd, hdc );

#if 0
	wsprintf( CommonBuffer,
		"IntLead: %d  ExtLead: %d  \n\
Height: %d  Width: %d  Ascent: %d  Descent: %d",
		tm.tmInternalLeading, tm.tmExternalLeading,
		tm.tmHeight, tm.tmAveCharWidth, tm.tmAscent, tm.tmDescent );
	MessageBox( ThisApphWnd, CommonBuffer, "FNT", MB_OK );
#endif

//	wsprintf( CommonBuffer, "MaxFonts: %d", MaxFonts );
//	MessageBox( ThisApphWnd, CommonBuffer, "Load Font", MB_OK );

	return TRUE;
} // End of InitChFont

