#ifndef _KXTYPE_H_
#define _KXTYPE_H_

#ifndef FAR
#define FAR far
#endif

#ifndef HUGE
#define HUGE huge
#endif

// Chinese encoding IDs
#define ASCII   0
#define BIG5    1
#define GB      2
#define HZ      3
#define INVALID_MODE -1

#define ROWS            50
#define KXROWS          10  // Max rows of kx translations buffer
#define KXDISP_ROWS     6  // Max Rows of kx translations can be shown
#define COLUMNS 80	// Default ideal number of columns
#define MAXCOL  80  // Maximum number of characters excluding the Null
                    // terminator you can actually stuff into a line.
					// This must be >= COLUMNS
#define DEF_LINESPACE	2

#define COLINC  81  // Actual size allocated for each line.

#define MAXLINES 32767 // Maximum number of lines can be read.
#define CBUFSIZE 512   // Common buffer size

#define BufRowPtr(r)    (r*COLINC)
#define NextRowPtr(b)   (b+COLINC)
#define PrevRowPtr(b)   (b-COLINC)

/* Screen attributes */
#define HZ_BYTE			128
#define HZ_BYTE1		(64|HZ_BYTE)
#define HZ_BYTE2		(32|HZ_BYTE)
#define ASCIIBYTE		16

#define NO_MODE  \
	((BYTE)(~HZ_BYTE1)&(BYTE)(~HZ_BYTE2)&(BYTE)(~ASCIIBYTE))
#define ALL_MODE ((BYTE)(~NO_MODE))

#define NoMode(b)			(b & NO_MODE)
#define HZByte(b)			(b & HZ_BYTE)
#define HZByte1(b)			(b & (HZ_BYTE1&(BYTE)(~HZ_BYTE)))
#define HZByte2(b)			(b & (HZ_BYTE2&(BYTE)(~HZ_BYTE)))

#define IsHZByte(b)			HZByte(b)
#define IsHZByte1(b)		HZByte1(b)
#define IsHZByte2(b)		HZByte2(b)
#define IsASCIIByte(b)		(b & ASCIIBYTE)

#define SetToNoMode(b)		(b = NoMode(b))
#define SetToHZByte(b)		(b = NoMode(b)|HZ_BYTE)
#define SetToHZByte1(b)		(b = NoMode(b)|HZ_BYTE1)
#define SetToHZByte2(b)		(b = NoMode(b)|HZ_BYTE2)
#define SetToASCIIByte(b)	(b = NoMode(b)|ASCIIBYTE)

#define CHAR_WIDTH(TW)      CharWidth
#define CCHAR_WIDTH(TW)     CCharWidth
#define CHAR_HEIGHT(TW)     CharHeight
#define ENCODING(TW)        Encoding
#define SHOW_GB_HZMARK(TW)  ShowGBHZMark
#define LINE_SPACE(TW)      LineSpace
#define LINE_HEIGHT(TW)     LineHeight
#define FONT_LEADING(TW)    FontLeading
#define LEFT_MARGIN(TW)     LeftMargin

#define UpdateTop(TW) TW->UpdateRC.top
#define UpdateBottom(TW) TW->UpdateRC.bottom
#define UpdateLeft(TW) TW->UpdateRC.left
#define UpdateRight(TW) TW->UpdateRC.right

#define TWFileValid(TW) (TW->AssocFile && TW->AssocFile->FilePtr)

typedef struct zFileInfo {
	char szFileName[512];	// Full path with name of the file being read
	FILE *FilePtr;		// Pointer to the file the user is reading
						// Changes if the user Open a new file.
	int  BookMarkLine;  // Line number of the bookmark
	long FileMark;		// Current mark for the file reading

    int  FileLine;      // Current line of the file user is reading
			// Changes if scroll, open file, and close file occur.
    long TotalLines;    // Number of lines the file has
			// Changes if the user opens a new file.

    HLOCAL hLineSize;   // An array of the size of every line in the file
	char *LineSize;		// Change if the user opens a new file.
} FileInfo;

typedef struct zTextWin {
/*
    // If implement to have file, buffer, and screen image:

	char *ScrImg;   // These are screen image data.  This is buffer
    int  TopRow;    // This is first row of display in buffer or file
    int  ImgHead;   // Use if implements circular buffer
    int  ImgRows;   // Max rows Img buffer can contain

    int  TotalRows;   // Total rows in buffer or file

*/
	char *Buf;
	HWND hWnd;
	WORD Width;
	WORD Height;
    int  x, y;
	int  Rows;	// Number of rows and columns that can fit
	int  Cols;	// in the current window.  Default 50, 80
				// and changes with the window size.
	int  ScrollRowLimit;	// Minimum # of lines must be visible
	int  ScrollColLimit;	// Changes with the window size.
	int  MaxRows;	// Maximum number of rows can contain.
	int  BufRows;  // Actual number of lines in the buffer.
					// Changes with window size and availability of text
					// to load from the file.
	int  BufTopRow; // First line of the buffer in memory.
				 // Will be used if I decide to implement circular buffer.
    int  TopRow;  // First row in display correspond to the row in buffer
    RECT UpdateRC;  // Rectangle need to be update
	BOOL ReloadBuffer;	// Indicate if needed to load buffer again
						// Changes if you scroll the text
    BOOL ScrollAble;    // When there is no file open or no associate file,
                        // this determine if the window can be scroll.
    FileInfo *AssocFile;
} TextWin;

typedef struct zChFontInfo
{
	char FileName[13];
    int  Encoding;
    BOOL JianTi;
    int  MaxFonts;
	int  Gap;
    char FontName[33];
    int  Size;
} ChFontInfo; 

#endif
