#include <stdio.h>
#include <string.h>
#define STRICT
#include <windows.h>
#include "kxtype.h"
#include "kxglobal.h"
#include "kxwproto.h"


#define COMP_MARKS 10

static long CompoundMarks[COMP_MARKS];
static int MarkIdx = 0;

static struct zErrText { char *str; } ReadErrStr[] = {
	"Success", "Not HZ encoding",
	"File not found", "Dict unsort",
	"File too big", 0  };

LPSTR PASCAL KXReadDictErrStr( long code )
{
	if ( code > 0 ) code = 0;
    else code = -code;
	return( (LPSTR)ReadErrStr[(int)code].str );
}

long PASCAL KXReadDict( BYTE HUGE *DictPtr, char *DictN )
{
    BYTE    *DN = (BYTE*)CommonBuffer;
	int 			I,
                    C,
                    TotalLine = 0;
	long			NChars = 0;
    char            AdjLines[2][20];
	int             Adj = 1, Chk = 0;
	FILE			*DictFilePtr;

    AdjLines[0][0] = 0;

	I = GetPrivateProfileString( "KuaiXue", "DictDir", 0,
		DN, CBUFSIZE, "kxwin.ini" );
	if ( I && DN[I]  != '\\' )
		strcat(DN, "\\" );
	strcat(DN, DictN);
    switch( Encoding ){
    case BIG5: strcat(DN, ".b5"); break;
    case GB:   strcat(DN, ".gb"); break;
    case HZ:   strcat(DN, ".hz"); break;
	default:   return -1L;
	}
	DictFilePtr = fopen(DN, "r");

    if (!DictFilePtr) return -2L;
	while (1) {
        if (fgets(DN, 512, DictFilePtr) == NULL) break;
		I = 0;
		Chk = 0;
		while (DN[I]) {
            if ( Chk == 0 ) {
                if ( DN[I] == ' ' ) {
                    Chk = 1;
                    AdjLines[Adj][I] = 0;
					if ( strcmp( AdjLines[Adj], AdjLines[1-Adj] ) < 0 ) {
						fclose(DictFilePtr);
						return -3L;
                    }
                } else
                    AdjLines[Adj][I] = DN[I];
			}
			DictPtr[NChars++] = (unsigned char) DN[I++];
			if (NChars >= MaxDictSize) {
				NChars = -4L;
				break;
			}
		}
        TotalLine ++;
		Adj = 1 - Adj;
	}
    fclose(DictFilePtr);
	return NChars;
}


static int  PASCAL CheckMatch(long J, BYTE HUGE * DictPtr)
{
	if (DictPtr[J] < ChnComp[CompIdx])
		return -1;
	if (DictPtr[J] > ChnComp[CompIdx])
		return 1;
	if (DictPtr[J + 1] < ChnComp[CompIdx + 1])
		return -1;
	if (DictPtr[J + 1] > ChnComp[CompIdx + 1])
		return 1;
	return 0;
}

static int  PASCAL LookUp(long *IPtr, BYTE HUGE * DictPtr, long NChars )
{
	unsigned char	C;
	int 			Cmp;
	long			I,
					Lo,
					Hi,
					Mid,
					OldMid;

	Lo = 0;
	Hi = NChars - 1;
	while (1) {
		OldMid = Mid;
		Mid = (Lo + Hi) / 2;
        if (Mid == OldMid)      Mid = Hi;
		I = Mid;
        while (DictPtr[I] != '\n' && I > 0) I--;
        if (I > 0) I++;
		/* check for a match */
        Cmp = CheckMatch(I+HzOff, DictPtr);

        /* Found */
		if (!Cmp) {
			*IPtr = I;
			return 1;
		}
        if (Mid == Hi)  return 0;
        if (Cmp < 0)    Lo = Mid;
        else            Hi = Mid;
	}
}

int  PASCAL KXLookUpSingle(long *IPtr, BYTE HUGE * DictPtr, long NChars )
{
	CompIdx = CLICKC_BYTE1_POS;
    return LookUp( IPtr, DictPtr, NChars );
}


#define CharEqual( I, Dict, Idx) (Dict[I] == ChnComp[Idx] && \
								  Dict[I+1] == ChnComp[Idx+1])
/*
 * Try to see if two compounds are equal.
 * Big5: space or '\0' signify the end of compound
 * HZ:   '~'
 */
static int    PASCAL CompoundCmp(long I)
{
	int 			CIdx,
					Matches = 0;

	CIdx = CompIdx;
	for (; CIdx < COMP_CCNUM * 2 + 1 && ChnComp[CompIdx]; CIdx += 2) {
        if (CompDict[I+HzOff] == ' ')      break;
        if (!CharEqual(I, CompDict, CIdx)) return 0;
		I += 2;
        if (I >= NCompDictChars)           break;
		Matches++;
	}
	if (Matches)
		return (CompDict[I+HzOff] == ' ' && CIdx > CLICKC_BYTE1_POS );
	else
		return 0;
}


/*
 * Find all entries, which have their first chars match the first char
 * of the compound, in the dictionary for a match of the compound
 * we are looking for.
 */
static int   PASCAL CompoundMatch(long I)
{
	long			J = I;
	int 			Found = 0;

	while (1) {
        while (J > 0 && CompDict[J] != '\n') J--;
		if (J > 0) J++;
        if (CharEqual(J+HzOff, CompDict, CompIdx)) {
			if (CompoundCmp(J+HzOff)) {
				CompoundMarks[MarkIdx++] = J;
                Found ++;
			}
			J -= 2;
		} else
			break;
        if ( J <= 0 ) break;
	}
	J = I;
	while (1) {
        while (J < NCompDictChars && CompDict[J] != '\n') J++;
		if (J < NCompDictChars - 1) J++;
		else						break;
        if (CharEqual(J+HzOff, CompDict, CompIdx)) {
			if (CompoundCmp(J+HzOff)) {
                CompoundMarks[MarkIdx++] = J;
                Found ++;
			}
			J += 2;
		} else
			break;
	}
	return Found;
}

int  PASCAL KXLookUpCompound( BYTE HUGE * DictPtr, long NChars )
{
	long			I;
    int             Found = 0;

    for( MarkIdx = 0; MarkIdx < COMP_MARKS; MarkIdx ++ )
        CompoundMarks[MarkIdx] = -1;

	MarkIdx = 0;
	for (CompIdx = 0; CompIdx <= CLICKC_BYTE1_POS; CompIdx += 2) {
        /* Search for the compound by comparing only the
         * first character first */
		if (ChnComp[CompIdx] && LookUp(&I, DictPtr, NChars))
            /* If first char matched, try to find a complete compound
             * which matches */
			Found += CompoundMatch(I);
	}
	MarkIdx = 0;
    return Found;
}

long PASCAL  KXNextFoundCompound( void )
{
    if ( MarkIdx < COMP_MARKS )
		return CompoundMarks[MarkIdx++];
	return -1L;
}

