#define STRICT      // Strict type checking
#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include "kxmenuid.h"	// Menu identifiers
#include "kxglobal.h"   // Global variables
#include "kxwproto.h"	// Function Prototypes

static TextWin *TW = 0;

BOOL PASCAL FileVertScroll( int Lines )
{
	int Scroll;
	FileInfo *FI = TW->AssocFile;

	if ( FI && FI->FilePtr ) {
    	if ( Lines > 0 ) Scroll = -SeekForward( TW, Lines );
    	else if ( Lines < 0 ) Scroll = SeekBackward( TW, -Lines );
		else return TRUE;

		if ( Scroll != 0 ) {
            Scroll *= LINE_HEIGHT(TW);
			ScrollWindowEx( TW->hWnd, 0, Scroll,
				0, 0, 0, 0, SW_INVALIDATE|SW_ERASE );
			// We need to load buffer again.  This must be set before
			// update window to be in effect.
			SetScrollPos( TW->hWnd, SB_VERT, FI->FileLine, TRUE );
			TW->ReloadBuffer = 1;
			UpdateWindow( TW->hWnd );
		} else if ( FI->FileLine != 0 &&
        	GetScrollPos( TW->hWnd, SB_VERT) != FI->TotalLines )
				SetScrollPos( TW->hWnd, SB_VERT, FI->TotalLines, TRUE );
        return TRUE;
	}
	return FALSE;
}

int PASCAL BufForward( TextWin *TW, int Lines )
{
	int LinesLeft = TW->BufRows - TW->TopRow - 1;

	if ( LinesLeft < TW->ScrollRowLimit ) return 0;

	LinesLeft -= Lines;
	if ( LinesLeft < TW->ScrollRowLimit )
		Lines = TW->BufRows - TW->TopRow - TW->ScrollRowLimit;

	if ( Lines > 0 ) {  
		TW->TopRow += Lines;
		if ( TW->TopRow > TW->BufRows )
			TW->TopRow = TW->BufRows - 1;
		return Lines;
	}
	return 0;
}

int PASCAL BufBackward( TextWin *TW, int Lines )
{
	if ( TW->TopRow < 1 ) return 0;

	if ( TW->TopRow - Lines < 0 )
		Lines = TW->TopRow;

	TW->TopRow -= Lines;

	return Lines;
}

void PASCAL VertScroll( int Lines )
{
	if ( TW->ScrollAble ) {
		if ( Lines > 0 ) Lines = -BufForward( TW, Lines );
		else if ( Lines < 0 ) Lines = BufBackward( TW, -Lines );
		else					return;
		if ( Lines != 0 ) {
        	KXGetChnCompReset();
            Lines *= LINE_HEIGHT(TW);
			ScrollWindowEx( TW->hWnd, 0, Lines,
				0, 0, 0, 0, SW_INVALIDATE|SW_ERASE );
			SetScrollPos( TW->hWnd, SB_VERT, TW->TopRow, TRUE );
			TW->ReloadBuffer = TRUE;
			UpdateWindow(TW->hWnd);
		} else if ( TW->TopRow != 0 &&
			GetScrollPos( TW->hWnd, SB_VERT ) != TW->BufRows )
				SetScrollPos( TW->hWnd, SB_VERT, TW->BufRows, TRUE );
	} else if ( FileVertScroll( Lines ) )
		KXGetChnCompReset();
}


void PASCAL RespondToVScroll( WPARAM wParam, LPARAM lParam )
{
    if ( !TW ) return;

	switch( wParam ) {
	case SB_LINEUP: VertScroll( -1 ); break;
	case SB_LINEDOWN: VertScroll( 1 ); break; 
    case SB_PAGEUP: VertScroll( -(TW->Rows-1)); break;
    case SB_PAGEDOWN: VertScroll( TW->Rows-1 ); break;
	case SB_THUMBTRACK:
		if ( !TrackScroll ) break;
	case SB_THUMBPOSITION:
        if ( TWFileValid(TW) )
			FileVertScroll( LOWORD(lParam)-TW->AssocFile->FileLine);
		else
			VertScroll( LOWORD(lParam)-TW->TopRow);
		break;
	default: break;
    }
}

void PASCAL DrawFontLines( TextWin *TW )
{
	HDC hdc;
	TEXTMETRIC tm;
	int RowIdx, ColIdx, PixRow;
	HPEN hpen;

	hpen = CreatePen(PS_SOLID, 1, RGB(255, 0, 0));

	hdc = GetDC( TW->hWnd );
	SetROP2( hdc, R2_NOT );
	SelectObject( hdc, ASCIIFont );
    SelectObject( hdc, hpen );
	GetTextMetrics( hdc, &tm );

	PixRow = LINE_SPACE(TW) - FONT_LEADING(TW);
	MoveTo( hdc, 0, PixRow );
	LineTo( hdc, 1000, PixRow );

    if ( tm.tmExternalLeading ) {
		PixRow += tm.tmExternalLeading;
		MoveTo( hdc, 0, PixRow );
		LineTo( hdc, 1000, PixRow );
	}

	MoveTo( hdc, 0, PixRow + tm.tmInternalLeading - 1 );
	LineTo( hdc, 1000, PixRow + tm.tmInternalLeading - 1);

	PixRow += tm.tmAscent - 1;
	MoveTo( hdc, 0, PixRow );
	LineTo( hdc, 1000, PixRow );

	PixRow += tm.tmDescent - 1;
	MoveTo( hdc, 0, PixRow );
	LineTo( hdc, 1000, PixRow );
	ReleaseDC( TW->hWnd, hdc );

	DeleteObject( hpen );

}

void PASCAL DrawHighLite( TextWin *TW )
{
	HDC hdc;

    hdc = GetDC( TW->hWnd ); 
	SetROP2( hdc, R2_NOT );
	Rectangle( hdc, LEFT_MARGIN(TW), LINE_SPACE(TW)-HZLeading,
		LEFT_MARGIN(TW)+30*CHAR_WIDTH(TW), LINE_HEIGHT(TW)-HZLeading);
	ReleaseDC( TW->hWnd, hdc );
}

void PASCAL RespondToKeyDown( WPARAM wParam )
{
    if ( !TW ) return;

	switch( wParam ) {
	case VK_UP: VertScroll( -1 ); break;
	case VK_DOWN: VertScroll( 1 ); break;
    case VK_PRIOR: VertScroll( -(TW->Rows-1) ); break; // Page up
	case VK_NEXT: VertScroll( TW->Rows-1 ); break; // page down
	case 0x41:
		DrawFontLines(&FileDisp);
		break;
	case 0x43:
		DrawHighLite(&FileDisp);
		break;
	case 0x42:
		FontGAP --;
		InvalidateRect( ThisApphWnd, NULL, TRUE );
		break;
	case 0x46: // 'f'
		FontGAP ++;
		InvalidateRect( ThisApphWnd, NULL, TRUE );
		break;
	case 0x44:
		wsprintf( CommonBuffer, "FontGAP: %d", FontGAP );
		MessageBox( NULL, CommonBuffer, "Dis", MB_OK );
		break;
	case 0x58: // 'r'
    	if ( CurrentFont )
			FontGAP = CurrentFont->Gap;
        break;
	default:
		break;
	}
} // End of RespondToKeyDown()

BOOL CALLBACK _export AboutDlgProc( HWND hDlg, UINT msg,
									WPARAM wParam, LPARAM lParam )
{
	HWND hControl;
	int w, h, x, y;
	RECT RC;

	switch( msg ) {
	case WM_INITDIALOG:
		hControl = GetDlgItem( hDlg, ABOUT_DLG_OK );
		SetFocus(hControl);
		GetWindowRect( hDlg, &RC );
		w = RC.right - RC.left + 1;
		h = RC.bottom - RC.top + 1;
		GetWindowRect( ThisApphWnd, &RC);
		x = (RC.right + RC.left + 1 - w)/2;
        y = (RC.bottom + RC.top + 1 - h)/2;
		SetWindowPos( hDlg, NULL, x, y,
			w, h, SWP_NOZORDER|SWP_NOACTIVATE);
		break;
	case WM_CLOSE:
		wParam = ABOUT_DLG_OK;
	case WM_COMMAND:
		if ( wParam == ABOUT_DLG_OK ) {
			EndDialog( hDlg, NULL );
			return TRUE;
		}
	default:
		break;
	}
	return FALSE;
}


void PASCAL ChangeFont( BOOL isJianTi, int Size )
{
	ChFontInfo oFInfo;

	if ( CurrentFont &&
			(CurrentFont->Size != Size ||
			CurrentFont->JianTi != isJianTi) ) {
		memcpy( &oFInfo, CurrentFont, sizeof(ChFontInfo) );
		oFInfo.Size = Size;
		oFInfo.JianTi = isJianTi;
		// Make sure the font info is there
		if ( ReadFontInfoFromIni( &oFInfo ) ) {  
			if ( InitChFont( &FileDisp, NULL ) ) {  // Unload current font
				// Get new font information
				memcpy( CurrentFont, &oFInfo, sizeof(ChFontInfo));
				// Load a new font.
				CurrentFont = 0;
				ChangeEncode( &FileDisp, Encoding );
			}
		}
	}
}

//
// Respond to menu selections.
// Return True if need to update the display
//
BOOL PASCAL RespondToMenu( WPARAM wParam )
{
	switch( wParam ) {
    case IDM_FILE_EXIT: Quit(); break;
    case IDM_FILE_OPEN: OpenNewFile(&FileDisp); break;
	case IDM_FILE_CLOSE:
		CloseFile(&FileDisp); GetBanner(&FileDisp); return TRUE;
	case IDM_VIEW_ASCII:ChangeEncode(TW, ASCII); return TRUE;
	case IDM_VIEW_BIG5: ChangeEncode(TW, BIG5); return TRUE;
	case IDM_VIEW_GB: ChangeEncode(TW, GB); return TRUE;
	case IDM_VIEW_HZ: ChangeEncode(TW, HZ); return TRUE;
    case IDM_VIEW_HZMARK: CheckGBHZMarkMenu(); return TRUE;
	case IDM_VIEW_TSCROLL: CheckTScrollMenu(); return FALSE;
	case IDM_FONT_16X16:
		if ( CurrentFont ) ChangeFont( CurrentFont->JianTi, 16 );
        break;
	case IDM_FONT_24X24:
		if ( CurrentFont ) ChangeFont( CurrentFont->JianTi, 24 );
		break;
	case IDM_VIEW_JIANTI:
		if ( CurrentFont ) ChangeFont( TRUE, CurrentFont->Size );
		break;
	case IDM_VIEW_FANTI:
		if ( CurrentFont ) ChangeFont( FALSE, CurrentFont->Size );
		break;
	case IDM_SAVE_BKMK:	SaveBookMark( TW->AssocFile, 1 ); break;
	case IDM_GOTO_BKMK:	GotoBookMark( TW->AssocFile ); break;
	case IDM_HELP_ABOUT: {
		DLGPROC lpAboutDlgProc;
		lpAboutDlgProc = (DLGPROC)MakeProcInstance(
						(FARPROC)AboutDlgProc, ThisApphInst );
		DialogBox( ThisApphInst, "KXAboutDlg", ThisApphWnd, lpAboutDlgProc );
		FreeProcInstance( (FARPROC)lpAboutDlgProc );
		break;
	}
	default:
		wsprintf(CommonBuffer, "MENU ID: %d", (int)wParam);
		MessageBox( ThisApphWnd, CommonBuffer, "KXWin", MB_OK );
		break;
	}
	return FALSE;
} // End of RespondToMenu()

void PASCAL GetWindowSize( void )
{
	int NewHeight, NewWidth, KXHeight;


	GetClientRect( ThisApphWnd, &MainText );
	// Reserve 3 pix lines for the separation line.
	NewHeight = MainText.bottom - MainText.top - 3;
	NewWidth = MainText.right - MainText.left;
	KXHeight = NewHeight/4;	/* use 1/4 of the entire display for KX */

    if ( KXHeight > (KXDISP_ROWS*LINE_HEIGHT(TW)))
        KXHeight = KXDISP_ROWS*LINE_HEIGHT(TW);

	NewHeight -= KXHeight;
	MainText.bottom = NewHeight + 1;

	if ( KXDisp.hWnd != (HWND)NULL ) {
		SetWindowPos( KXDisp.hWnd, NULL, 0, NewHeight+2,
			NewWidth, KXHeight, SWP_NOZORDER|SWP_NOACTIVATE);
	}
	if ( FileDisp.hWnd != (HWND)NULL ) {
		SetWindowPos( FileDisp.hWnd, NULL, 0, 0,
			NewWidth, NewHeight-1, SWP_NOZORDER|SWP_NOACTIVATE);
	}
} // End of GetWindowSize


//
// Message receiver
//
LRESULT CALLBACK _export KXWndProc( HWND hWnd,
                                    UINT msg,
                                    WPARAM wParam,
                                    LPARAM lParam )
{
	HDC hdc;
	PAINTSTRUCT ps;

	switch( msg ) {
	case WM_PAINT:
		hdc = BeginPaint( ThisApphWnd, &ps );
		MoveTo( hdc, MainText.left, MainText.bottom-1 );
        LineTo( hdc, MainText.right, MainText.bottom-1 );
		EndPaint( ThisApphWnd, &ps );
		break;
	case WM_SIZE:
		if ( wParam == SIZE_MAXIMIZED || wParam == SIZE_RESTORED ) {
			GetWindowSize();
		} 
		break;
	case WM_DESTROY:
		PostQuitMessage( 0 );
        InitChFont(TW, NULL);         // so this will deinit the fonts loaded
		break;
	case WM_KEYDOWN:
        TW = &FileDisp;
		RespondToKeyDown( wParam ); break;
	case WM_CLOSE:	Quit(); break;
	case WM_COMMAND:
		TW = &FileDisp;
		if ( RespondToMenu( wParam ) )
			InvalidateRect( hWnd, NULL, FALSE );	// Refresh screen
		break;
	default:
        return( DefWindowProc( hWnd, msg, wParam, lParam) );
	}
	return 0L;
}  // End of KXWndProc

static void PASCAL CalcTextDim( TextWin *TW )
{
	RECT RC;

	GetClientRect( TW->hWnd, &RC );

	TW->Width = RC.right - RC.left;
	TW->Height = RC.bottom - RC.top;
    TW->Cols = (TW->Width-LEFT_MARGIN(TW))/CHAR_WIDTH(TW)+1;
    TW->Rows = TW->Height/LINE_HEIGHT(TW);

	TW->ScrollRowLimit = TW->Rows/2;
	TW->ScrollRowLimit += TW->ScrollRowLimit/2;
	if ( TW->ScrollRowLimit < 1 ) TW->ScrollRowLimit = 1;
	TW->ScrollColLimit = TW->Cols/2;
}


void PASCAL AdjustTextWin( TextWin *TW, BOOL Resized )
{
	if ( Resized ) CalcTextDim(TW);

    if ( !TWFileValid(TW) ) {
        if ( TW->Rows < TW->BufRows ) {
            TW->ScrollAble = TRUE;
            if ( TW->BufRows - TW->TopRow < TW->ScrollRowLimit )
                TW->TopRow = TW->BufRows - TW->ScrollRowLimit;
			// Enable Vertical Scrollbar
			ShowScrollBar( TW->hWnd, SB_VERT, TRUE );
			EnableScrollBar( TW->hWnd, SB_VERT, ESB_ENABLE_BOTH);
		} else {
            TW->TopRow = 0;
            TW->ScrollAble = FALSE;
            // Disable Vertical Scrolbar
			EnableScrollBar( TW->hWnd, SB_VERT, ESB_DISABLE_BOTH );
		}
		SetScrollPos( TW->hWnd, SB_VERT, TW->TopRow, TRUE );
	} else {
    	FileInfo *FI = TW->AssocFile;
        TW->ReloadBuffer = Resized;
		if( FI->TotalLines >= TW->Rows ) {
			if ( FI->TotalLines - FI->FileLine < TW->ScrollRowLimit )
				FileVertScroll( FI->TotalLines - TW->ScrollRowLimit
					- FI->FileLine );
			// Enable Vertical Scrollbar
			EnableScrollBar( TW->hWnd, SB_VERT, ESB_ENABLE_BOTH);
		} else {
			FileVertScroll( 0-TW->AssocFile->FileLine );
			// Disable Vertical Scrollbar
			EnableScrollBar( TW->hWnd, SB_VERT, ESB_DISABLE_BOTH );
		}
	}
}


#if 0
LPSTR ToBin( BYTE Sqr, int l )
{
    static char s[10][10];
    static char D[] = "Hota===";
    int i;
    BYTE S = 128;
    LPSTR sp = (LPSTR)s[l];

    for( i = 0; i < 8; i ++ ) {
        if ( Sqr & S )
            sp[i] = D[i];
        else
            sp[i] = '-';
        S >>= 1;
    }
    sp[i] = 0;
    return sp;
}
#endif


static int ReadDict( TextWin *TW, int *Row, 
		BYTE HUGE *DictPtr, LPSTR DictName, long *NChars )
{ 
	long NC;
	int OldRow = *Row;

	PrintToBuffer( TW, *Row, 0,
		"Wait... Loading dictionary.", TW_WRAPNON );
	NC = KXReadDict( DictPtr, (char*)DictName );
	if ( NC < 0L ) {
		wsprintf(CommonBuffer,
			"Load dictionary file %s<.encode>: Failed! %s.",
				DictName, KXReadDictErrStr(NC) );
		(*Row) += PrintToBuffer( TW, (*Row), 0,
				CommonBuffer, TW_WRAPNON );
		return 0;
	}
	*NChars = NC;
	EraseBuffer( &KXDisp, OldRow, (*Row)-1 );
	*Row = OldRow;
	return 1;
}

void PASCAL DoLookup( TextWin *TW, LPARAM lParam )
{
		int Row, Col, I;

		Row = (HIWORD(lParam)-LINE_SPACE(TW))/LINE_HEIGHT(TW);
		Col = (LOWORD(lParam)-LEFT_MARGIN(TW))/CHAR_WIDTH(TW);

		if ( !TWFileValid(TW) ) Row += TW->TopRow;

		if ( KXGetChnComp(TW, Row, Col) ) {
			int KXRow = 0;
			long FoundMark;

			EraseBuffer( &KXDisp, 0, KXDisp.BufRows-1 ); 
			for( Col = 0; !ChnComp[Col]; Col++ );

			if ( InitDictType != Encoding ) {
				if ( ReadDict( &KXDisp, &KXRow, Dict,
						"canman", &NDictChars )) 
					InitDictType = Encoding;
				else goto _KX_LookUp_Compound;
            }  
			if ( KXLookUpSingle( &FoundMark, Dict, NDictChars )) {
				I = 0;
				while( Dict[FoundMark] != '\n' ) {
					CommonBuffer[I++] = Dict[FoundMark++];
				}
				CommonBuffer[I] = 0;
			} else {
				wsprintf( CommonBuffer, "  Not found: %s%c%c%s",
					HZOnMark, ChnComp[CLICKC_BYTE1_POS],
					ChnComp[CLICKC_BYTE1_POS+1],	HZOffMark );
			}
            KXRow += PrintToBuffer( &KXDisp, KXRow, 0, CommonBuffer,
                TW_WRAPWIN );
            PrintToShengZi(CommonBuffer, "\n");
_KX_LookUp_Compound:
			if ( InitCompDictType != Encoding ) {
				if ( ReadDict( &KXDisp, &KXRow, CompDict,
						"compound", &NCompDictChars ))
					InitCompDictType = Encoding;
				else
					return;
			}
            if ( KXLookUpCompound( CompDict, NCompDictChars )) {
                while( (FoundMark = KXNextFoundCompound()) >= 0L ) {
                    I = 0;
                    while( CompDict[FoundMark] != '\n' ) {
                        CommonBuffer[I++] = CompDict[FoundMark++];
					}
					CommonBuffer[I] = 0;
                    KXRow += PrintToBuffer( &KXDisp, KXRow, 0, CommonBuffer,
                        TW_WRAPWIN );
                    PrintToShengZi(CommonBuffer, "\n");
                }
            } else {
				lstrcpy( CommonBuffer, "  No compound found.");
				PrintToBuffer( &KXDisp, KXRow, 0, CommonBuffer,
					TW_WRAPWIN );
			}
		}
}


LRESULT CALLBACK _export KXTextWndProc( HWND hWnd,
                                    UINT msg,
                                    WPARAM wParam,
                                    LPARAM lParam )
{
	static HWND MW = 0; /* mouse pressed window */

	TW = ( hWnd == FileDisp.hWnd ? &FileDisp : &KXDisp );
	switch( msg ) {
    case WM_PAINT:
		if ( GetUpdateRect(hWnd, &TW->UpdateRC, FALSE) )
			RefreshScreen(TW);
		break;
	case WM_LBUTTONDOWN: MW = hWnd;	break;
	case WM_LBUTTONUP:
		if ( MW == hWnd ) DoLookup(TW, lParam); break;
	case WM_SIZE:
		if ( wParam == SIZE_MAXIMIZED || wParam == SIZE_RESTORED ) {
			AdjustTextWin(TW, TRUE);
		}
		break;
	case WM_VSCROLL:RespondToVScroll( wParam, lParam ); break;
    default:  return DefWindowProc( hWnd, msg, wParam, lParam );
    }
    return 0L;
}

//
// End of KXWin.c
//
